@@ -11,6 +11,19 @@ ISL_SEM_TO_SEM = {
     "__isl_keep": SEM_KEEP,
     }
 
+NON_COPYABLE = ["ctx", "printer", "access_info"]
+NON_COPYABLE_WITH_ISL_PREFIX = ["isl_"+i for i in NON_COPYABLE]
+
+PYTHON_RESERVED_WORDS = """
+and       del       from      not       while
+as        elif      global    or        with
+assert    else      if        pass      yield
+break     except    import    print
+class     exec      in        raise
+continue  finally   is        return
+def       for       lambda    try
+""".split()
+
 class Argument:
     def __init__(self, name, semantics, base_type, ptr):
         self.name = name
@@ -52,7 +65,8 @@ class Method:
                 and self.return_ptr == "*" == self.args[0].ptr
                 and self.return_base_type == self.args[0].base_type
                 and self.return_semantics is SEM_GIVE
-                and not self.mutator_veto)
+                and not self.mutator_veto
+                and self.args[0].base_type in NON_COPYABLE_WITH_ISL_PREFIX)
 
     def __repr__(self):
         return "<method %s>" % self.c_name
@@ -346,6 +360,9 @@ class FunctionData:
             print "SKIP:", cls, name
             return
 
+        if name in PYTHON_RESERVED_WORDS:
+            name = name + "_"
+
         words = return_base_type.split()
         return_semantics, words = filter_semantics(words)
         words = [w for w in words if w not in ["struct", "enum"]]
@@ -487,6 +504,8 @@ def write_wrapper(outf, meth):
                 passed_args.append(arg.name)
             input_args.append("%s *%s" % (arg.base_type, arg.name))
 
+            docs.append(":param %s: string" % arg.name)
+
         elif arg.base_type == "int" and arg.ptr == "*":
             if arg.name in ["exact", "tight"]:
                 body.append("int arg_%s;" % arg.name)
@@ -500,9 +519,15 @@ def write_wrapper(outf, meth):
         elif arg.base_type == "isl_int" and arg.ptr == "*":
             # assume it's meant as a return value
             body.append("""
-                py::object arg_%s(py::handle<>((PyObject *) Pympz_new()));
-                """ % arg.name)
-            passed_args.append("&(Pympz_AS_MPZ(arg_%s.ptr()))" % arg.name)
+                py::object arg_%(name)s(py::handle<>((PyObject *) Pympz_new()));
+                managed_int arg_mi_%(name)s;
+
+                """ % dict(name=arg.name))
+            passed_args.append("&arg_mi_%s.m_data" % arg.name)
+            post_call.append("""
+                isl_int_get_gmp(arg_mi_%(name)s.m_data, Pympz_AS_MPZ(arg_%(name)s.ptr()));
+                """ % dict(name=arg.name))
+
             extra_ret_vals.append("arg_%s" % arg.name)
             extra_ret_descrs.append("%s (integer)" % arg.name)
 
@@ -511,20 +536,23 @@ def write_wrapper(outf, meth):
         elif arg.base_type == "isl_int" and not arg.ptr:
             input_args.append("py::object %s" % ("arg_"+arg.name))
             checks.append("""
-                py::handle<> converted_arg_%(name)s;
+                managed_int arg_mi_%(name)s;
                 {
                   PyObject *converted;
                   if (Pympz_convert_arg(arg_%(name)s.ptr(), &converted) == 0)
                     throw py::error_already_set();
-                  converted_arg_%(name)s = py::handle<>(converted);
+                  py::handle<> converted_arg_%(name)s = py::handle<>(converted);
+                  isl_int_set_gmp(arg_mi_%(name)s.m_data,
+                    Pympz_AS_MPZ(converted_arg_%(name)s.get()));
                 }
                 """ % dict(name=arg.name, meth="%s_%s" % (meth.cls, meth.name)))
-            passed_args.append("Pympz_AS_MPZ(converted_arg_%s.get())" % arg.name)
+            passed_args.append("arg_mi_%s.m_data" % arg.name)
 
             docs.append(":param %s: integer" % arg.name)
 
         elif arg.base_type.startswith("isl_") and arg.ptr == "*":
             arg_cls = arg.base_type[4:]
+            arg_descr = ":param %s: :class:`%s`" % (arg.name, to_py_class(arg_cls))
 
             if arg.semantics is None and arg.base_type != "isl_ctx":
                 raise Undocumented(meth)
@@ -534,19 +562,42 @@ def write_wrapper(outf, meth):
                   PYTHON_ERROR(ValueError, "passed invalid arg to isl_%(meth)s for %(name)s");
                 """ % dict(name=arg.name, meth="%s_%s" % (meth.cls, meth.name)))
 
+            copyable = arg_cls not in NON_COPYABLE
             if arg.semantics is SEM_TAKE:
-                if not (arg_idx == 0 and meth.is_mutator):
-                    post_call.append("arg_%s->invalidate();" % arg.name)
+                if copyable:
+                    checks.append("""
+                        if (!arg_%(name)s || !arg_%(name)s->is_valid())
+                          PYTHON_ERROR(ValueError, "passed invalid arg to isl_%(meth)s for %(name)s");
+                        std::auto_ptr<%(cls)s> auto_arg_%(name)s;
+                        {
+                            isl_%(cls)s *tmp_ptr = isl_%(cls)s_copy(arg_%(name)s->m_data);
+                            if (!tmp_ptr)
+                                throw std::runtime_error("failed to copy arg %(name)s on entry to %(meth)s");
+                            auto_arg_%(name)s = std::auto_ptr<%(cls)s>(new %(cls)s(tmp_ptr));
+                        }
+                        """ % dict(
+                            name=arg.name,
+                            meth="%s_%s" % (meth.cls, meth.name),
+                            cls=arg_cls))
+
+                    post_call.append("auto_arg_%s.release();" % arg.name)
+                    passed_args.append("auto_arg_%s->m_data" % arg.name)
+
+                else:
+                    if not (arg_idx == 0 and meth.is_mutator):
+                        post_call.append("arg_%s->invalidate();" % arg.name)
+
+                    passed_args.append("arg_%s->m_data" % arg.name)
+
+                    if arg_idx == 0 and meth.is_mutator:
+                        arg_descr += " (mutated in-place)"
+                    else:
+                        arg_descr += " (:ref:`becomes invalid <auto-invalidation>`)"
+            else:
+                passed_args.append("arg_%s->m_data" % arg.name)
 
-            passed_args.append("arg_%s->m_data" % arg.name)
             input_args.append("%s *%s" % (arg_cls, "arg_"+arg.name))
 
-            arg_descr = ":param %s: :class:`%s`" % (arg.name, to_py_class(arg_cls))
-            if arg.semantics is SEM_TAKE:
-                if arg_idx == 0 and meth.is_mutator:
-                    arg_descr += " (mutated in-place)"
-                else:
-                    arg_descr += " (:ref:`becomes invalid <auto-invalidation>`)"
             docs.append(arg_descr)
 
         elif arg.base_type.startswith("isl_") and arg.ptr == "**":
@@ -741,6 +792,14 @@ def write_exposer(outf, meth, arg_names, doc_str, static_decls):
     if meth.name == "size" and len(meth.args) == 1:
         py_name = "__len__"
 
+    if meth.name == "get_hash" and len(meth.args) == 1:
+        py_name = "__hash__"
+
+    extra_py_names = []
+
+    if meth.name == "is_equal":
+        extra_py_names.append("__eq__")
+
     #if meth.is_static:
         #doc_str = "(static method)\n" + doc_str
 
@@ -751,11 +810,13 @@ def write_exposer(outf, meth, arg_names, doc_str, static_decls):
         extra_stuff = extra_stuff+", py::return_self<>()"
 
     wrap_class = CLASS_MAP.get(meth.cls, meth.cls)
-    outf.write("wrap_%s.def(\"%s\", %s%s);\n" % (
-        wrap_class, py_name, func_name, extra_stuff))
-    if meth.is_static:
-        static_decls.append("wrap_%s.staticmethod(\"%s\");\n" % (
-            wrap_class, py_name))
+
+    for exp_py_name in [py_name]+extra_py_names:
+        outf.write("wrap_%s.def(\"%s\", %s%s);\n" % (
+            wrap_class, exp_py_name, func_name, extra_stuff))
+        if meth.is_static:
+            static_decls.append("wrap_%s.staticmethod(\"%s\");\n" % (
+                wrap_class, exp_py_name))
 
 
 
